@@ -2,7 +2,7 @@
  * wiggle - apply rejected patches
  *
  * Copyright (C) 2005 Neil Brown <neilb@cse.unsw.edu.au>
- * Copyright (C) 2010 Neil Brown <neilb@suse.de>
+ * Copyright (C) 2010-2013 Neil Brown <neilb@suse.de>
  *
  *
  *    This program is free software; you can redistribute it and/or modify
@@ -16,14 +16,14 @@
  *    GNU General Public License for more details.
  *
  *    You should have received a copy of the GNU General Public License
- *    along with this program; if not, write to the Free Software Foundation, Inc.,
- *    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
+ *    along with this program.
  *
  *    Author: Neil Brown
  *    Email: <neilb@suse.de>
  */
 
 #include "wiggle.h"
+#include <limits.h>
 
 /*
  * Second attempt at merging....
@@ -77,7 +77,7 @@ static int check_alreadyapplied(struct file af, struct file cf,
  * the change the mode of display - between displaying the merger
  * and displaying the separate streams.
  * A 'conflict' can only be displayed as separate stream so when
- * one is found, we need to find a preceeding and trailing cut-point
+ * one is found, we need to find a preceding and trailing cut-point
  * and enlarge the conflict to that range.
  * A suitable location is one where all three streams are at a line-end.
  */
@@ -89,32 +89,40 @@ static int is_cutpoint(struct merge m,
 		(m.c == 0 || ends_line(cf.list[m.c-1])));
 }
 
-static int isolate_conflicts(struct file af, struct file bf, struct file cf,
-			     struct csl *csl1, struct csl *csl2, int words,
-			     struct merge *m, int show_wiggles)
+int isolate_conflicts(struct file af, struct file bf, struct file cf,
+		      struct csl *csl1, struct csl *csl2, int words,
+		      struct merge *m, int show_wiggles,
+		      int *wigglesp)
 {
-	/* A conflict indicates that something is definitely wrong
+	/* A Conflict indicates that something is definitely wrong
 	 * and so we need to be a bit suspicious of nearby apparent matches.
-	 * To display a conflict effectively we expands it's effect to
+	 * To display a conflict effectively we expands its effect to
 	 * include any Extraneous, Unmatched, Changed or AlreadyApplied text.
 	 * Also, unless 'words', we need to include any partial lines
 	 * in the Unchanged text that forms the border of a conflict.
 	 *
 	 * A Changed text may also border a conflict, but it can
 	 * only border one conflict (where as an Unchanged can border
-	 * a preceeding and a following conflict).
+	 * a preceding and a following conflict).
 	 * The 'new' section of a Changed text appears in the
 	 * conflict as does any part of the original before
 	 * a newline.
 	 *
-	 * If 'show_wiggles' is set we treat wiggles like conflicts.
-	 * A 'wiggle' is implied by any Extraneous text being ignored,
-	 * or any line that has both Changed and Unmatched content.
-	 * (Unmatched content where nothing is changed is common and not
-	 *  really a 'wiggle').
+	 * A hunk header (Extraneous) is never considered part of a
+	 * conflict.  It thereby can serve as a separator between
+	 * conflicts.
 	 *
-	 * A hunk header is never considered part of a conflict.  It
-	 * thereby can serve as a separator between conflicts.
+	 * Extended conflicts are marked by setting ->in_conflict in
+	 * the "struct merge".  This is '1' for an Unchanged, Changed,
+	 * or (Extraneous) hunk header which borders the conflict,
+	 * '2' for a merger which is truly in conflict, and '3' for
+	 * a merger which is causing a 'wiggle'.
+	 * When in_conflict == 1, the 'lo' and 'hi' fields indicate
+	 * how much of the 'a' file is included in the conflict, the rest
+	 * being part of the clean result.
+	 * Elements in af from m->a to m->a+m->lo are in the preceding
+	 * conflict, from m->a+m->lo to m->a+m->hi are clean, and
+	 * m->a+m->hi to m->a+m->al are in the following conflict.
 	 *
 	 * We need to ensure there is adequate context for the conflict.
 	 * So ensure there are at least 3 newlines in Extraneous or
@@ -122,21 +130,45 @@ static int isolate_conflicts(struct file af, struct file bf, struct file cf,
 	 * as including a hunk header.
 	 * If there are 3, and they are all in 'Unchanged' sections, then
 	 * that much context is not really needed - reduce it a bit.
+	 *
+	 * If a wiggle is adjacent to a conflict then:
+	 * - if show_wiggles is set, we just merge them
+	 * - if it is not set, we still want to count the wiggle.
 	 */
 	int i, j, k;
-	int cnt = 0;
+	int cnt = 0, wiggles = 0;
+	int in_wiggle = 0;
 	int changed = 0;
 	int unmatched = 0;
+	int extraneous = 0;
+
+	for (i = 0; m[i].type != End; i++)
+		m[i].in_conflict = 0;
 
 	for (i = 0; m[i].type != End; i++) {
+		/* The '3' here is a count of newlines.  Once we find
+		 * that many newlines of the particular type, we have escaped.
+		 */
 		if (m[i].type == Changed)
-			changed = 1;
+			changed = 3;
 		if (m[i].type == Unmatched)
-			unmatched = 1;
-		if (m[i].type == Conflict ||
-		    (show_wiggles && ((changed && unmatched)
-					|| m[i].type == Extraneous))) {
-			/* We have a conflict (or wiggle) here.
+			unmatched = 3;
+		if (m[i].type == Extraneous && bf.list[m[i].b].start[0])
+			/* hunk headers don't imply wiggles, other
+			 * extraneous text does.
+			 */
+			extraneous = 3;
+
+		if (m[i].type != Unchanged && changed && (unmatched || extraneous)) {
+			if (!in_wiggle)
+				wiggles++;
+			in_wiggle = 1;
+		} else
+			in_wiggle = 0;
+
+		if ((m[i].type == Conflict) ||
+		    (show_wiggles && in_wiggle)) {
+			/* We have a conflict or wiggle here.
 			 * First search backwards for an Unchanged marking
 			 * things as in_conflict.  Then find the
 			 * cut-point in the Unchanged.  If there isn't one,
@@ -145,73 +177,84 @@ static int isolate_conflicts(struct file af, struct file bf, struct file cf,
 			 * Then search forward doing the same thing.
 			 */
 			int newlines = 0;
-			cnt++;
-			m[i].in_conflict = 1;
+			m[i].in_conflict = m[i].type == Conflict ? 2 : 3;
 			j = i;
 			while (--j >= 0) {
+				int firstk;
+
 				if (m[j].type == Extraneous &&
 				    bf.list[m[j].b].start[0] == '\0')
 					/* hunk header - not conflict any more */
 					break;
+				if (m[j].in_conflict > 1)
+					/* Merge the conflicts */
+					break;
 				if (!m[j].in_conflict) {
 					m[j].in_conflict = 1;
 					m[j].lo = 0;
-				} else if (m[j].type == Changed) {
-					/* This can no longer form a border */
-					m[j].hi = -1;
-					/* We merge these conflicts and stop searching */
-					cnt--;
-					break;
 				}
+				/* Following must set m[j].hi, or set
+				 * in_conflict > 1
+				 */
 				if (m[j].type == Extraneous) {
 					for (k = m[j].bl; k > 0; k--)
 						if (ends_line(bf.list[m[j].b+k-1]))
 							newlines++;
 				}
 
-				if (m[j].type == Unchanged || m[j].type == Changed) {
-					/* If we find enough newlines in this section,
-					 * then we only really need 1, but would rather
-					 * it wasn't the first one.  'firstk' allows us
-					 * to track which newline we actually use
-					 */
-					int firstk = m[j].al+1;
-					if (words) {
-						m[j].hi = m[j].al;
-						break;
-					}
-					/* need to find the last line-break, which
-					 * might be after the last newline, if there
-					 * is one, or might be at the start
-					 */
-					for (k = m[j].al; k > 0; k--)
-						if (ends_line(af.list[m[j].a+k-1])) {
-							if (firstk >= m[j].al)
-								firstk = k;
-							newlines++;
-							if (newlines >= 3) {
-								k = firstk;
-								break;
-							}
-						}
-					if (k > 0)
-						m[j].hi = k;
-					else if (is_cutpoint(m[j], af,bf,cf))
-						m[j].hi = 0;
+				if (m[j].type != Unchanged &&
+				    m[j].type != Changed) {
+					if (m[j].type == Conflict)
+						m[j].in_conflict = 2;
 					else
-						/* no start-of-line found... */
-						m[j].hi = -1;
-					if (m[j].hi > 0 && m[j].type == Changed) {
-						/* this can only work if start is
-						 * also a line break */
-						if (is_cutpoint(m[j], af,bf,cf))
-							/* ok */;
-						else
-							m[j].hi = -1;
-					}
-					if (m[j].hi >= 0)
-						break;
+						m[j].in_conflict = m[i].in_conflict;
+					continue;
 				}
+				/* If we find enough newlines in this section,
+				 * then we only really need 1, but would rather
+				 * it wasn't the first one.  'firstk' allows us
+				 * to track which newline we actually use
+				 */
+				firstk = m[j].al+1;
+				if (words) {
+					m[j].hi = m[j].al;
+					break;
+				}
+				/* need to find the last line-break, which
+				 * might be after the last newline, if there
+				 * is one, or might be at the start
+				 */
+				for (k = m[j].al; k > 0; k--)
+					if (ends_line(af.list[m[j].a+k-1])) {
+						if (firstk > m[j].al)
+							firstk = k;
+						newlines++;
+						if (newlines >= 3) {
+							k = firstk;
+							break;
+						}
+					}
+				if (k > 0)
+					m[j].hi = k;
+				else if (j == 0)
+					m[j].hi = firstk;
+				else if (is_cutpoint(m[j], af,bf,cf))
+					m[j].hi = 0;
+				else
+					/* no start-of-line found... */
+					m[j].hi = -1;
+				if (m[j].hi > 0 &&
+				    (m[j].type == Changed)) {
+					/* this can only work if start is
+					 * also a line break */
+					if (is_cutpoint(m[j], af,bf,cf))
+						/* ok */;
+					else
+						m[j].hi = -1;
+				}
+				if (m[j].hi >= 0)
+					break;
+				m[j].in_conflict = m[i].in_conflict;
 			}
 
 			/* now the forward search */
@@ -221,79 +264,159 @@ static int isolate_conflicts(struct file af, struct file bf, struct file cf,
 				    bf.list[m[j].b].start[0] == '\0')
 					/* hunk header - not conflict any more */
 					break;
-				m[j].in_conflict = 1;
 				if (m[j].type == Extraneous) {
 					for (k = 0; k < m[j].bl; k++)
 						if (ends_line(bf.list[m[j].b+k]))
 							newlines++;
 				}
-				if (m[j].type == Unchanged || m[j].type == Changed) {
-					m[j].hi = m[j].al;
-					if (words) {
-						m[j].lo = 0;
-						break;
-					}
-					/* need to find a line-break, which might be at
-					 * the very beginning, or might be after the
-					 * first newline - if there is one
-					 */
-					if (is_cutpoint(m[j], af,bf,cf))
-						m[j].lo = 0;
-					else {
-						/* If we find enough newlines in this section,
-						 * then we only really need 1, but would rather
-						 * it wasn't the first one.  'firstk' allows us
-						 * to track which newline we actually use
-						 */
-						int firstk = -1;
-						for (k = 0 ; k < m[j].al ; k++)
-							if (ends_line(af.list[m[j].a+k])) {
-								if (firstk <= 0)
-									firstk = k;
-								newlines++;
-								if (newlines >= 3) {
-									k = firstk;
-									break;
-								}
-							}
-						if (firstk >= 0 &&
-						    m[j+1].type == Unmatched) {
-							/* If this Unmatched exceeds 3 lines, just stop here */
-							int p;
-							int nl = 0;
-							for (p = 0; p < m[j+1].al ; p++)
-								if (ends_line(af.list[m[j+1].a+p])) {
-									nl++;
-									if (nl > 3)
-										break;
-								}
-							if (nl > 3)
-								k = firstk;
-						}
-						if (k < m[j].al)
-							m[j].lo = k+1;
-						else
-							/* no start-of-line found */
-							m[j].lo = m[j].al+1;
-					}
-					if (m[j].lo <= m[j].al+1 && m[j].type == Changed) {
-						/* this can only work if the end is a line break */
-						if (is_cutpoint(m[j+1], af,bf,cf))
-							/* ok */;
-						else
-							m[j].lo = m[j].al+1;
-					}
-					if (m[j].lo < m[j].al+1)
-						break;
+				if (m[j].type != Unchanged &&
+				    m[j].type != Changed) {
+					if (m[j].type == Conflict)
+						m[j].in_conflict = 2;
+					else
+						m[j].in_conflict = m[i].in_conflict;
+					continue;
 				}
+				m[j].in_conflict = 1;
+				m[j].hi = m[j].al;
+				if (words) {
+					m[j].lo = 0;
+					break;
+				}
+				/* need to find a line-break, which might be at
+				 * the very beginning, or might be after the
+				 * first newline - if there is one
+				 */
+				if (is_cutpoint(m[j], af,bf,cf))
+					m[j].lo = 0;
+				else {
+					/* If we find enough newlines in this section,
+					 * then we only really need 1, but would rather
+					 * it wasn't the first one.  'firstk' allows us
+					 * to track which newline we actually use
+					 */
+					int firstk = -1;
+					for (k = 0 ; k < m[j].al ; k++)
+						if (ends_line(af.list[m[j].a+k])) {
+							if (firstk < 0)
+								firstk = k;
+							newlines++;
+							if (newlines >= 3) {
+								k = firstk;
+								break;
+							}
+						}
+					if (newlines < 3 &&
+					    m[j+1].type  == End)
+						/* Hit end of file, pretend we found 3 newlines. */
+						k = firstk;
+
+					if (firstk >= 0 &&
+					    m[j+1].type == Unmatched) {
+						/* If this Unmatched exceeds 3 lines, just stop here */
+						int p;
+						int nl = 0;
+						for (p = 0; p < m[j+1].al ; p++)
+							if (ends_line(af.list[m[j+1].a+p])) {
+								nl++;
+								if (nl > 3)
+									break;
+							}
+						if (nl > 3)
+							k = firstk;
+					}
+					if (k < m[j].al)
+						m[j].lo = k+1;
+					else
+						/* no start-of-line found */
+						m[j].lo = m[j].al+1;
+				}
+				if (m[j].lo <= m[j].al+1 &&
+				    (m[j].type == Changed)) {
+					/* this can only work if the end is a line break */
+					if (is_cutpoint(m[j+1], af,bf,cf))
+						/* ok */;
+					else
+						m[j].lo = m[j].al+1;
+				}
+				if (m[j].lo < m[j].al+1)
+					break;
+				m[j].in_conflict = m[i].in_conflict;
 			}
-			i = j - 1;
-		}
-		if (m[i].al > 0 && ends_line(af.list[m[i].a+m[i].al-1])) {
-			unmatched = 0;
-			changed = 0;
+			if (m[j-1].in_conflict == 1)
+				i = j - 1;
+			else
+				/* A hunk header bordered the conflict */
+				i = j;
+
+			/* If any of the merges are Changed or Conflict,
+			 * then this really is a Conflict or Wiggle.
+			 * If not they are just Unchanged, Unmatched,
+			 * Extraneous or AlreadyApplied, and so don't
+			 * really count.
+			 * Note that the first/last merges (in_conflict==1)
+			 * can be Changed and so much be check separately.
+			 */
+			if (m[j].type == Changed)
+				goto out;
+			for (j = i-1; j >= 0 && m[j].in_conflict > 1; j--)
+				if (m[j].type == Changed || m[j].type == Conflict)
+					goto out;
+			if (j >= 0 && m[j].type == Changed)
+				goto out;
+			/* False alarm, no real conflict/wiggle here as
+			 * nothing changed. */
+			if (j < 0)
+				j = 0;
+			if (m[j].in_conflict == 1) {
+				m[j].hi = m[j].al;
+				if (m[j].lo == 0)
+					m[j].in_conflict = 0;
+				j++;
+			}
+			while (j <= i)
+				m[j++].in_conflict = 0;
+		out:
+			if (m[i].type == End)
+				break;
 		}
+		for (k = 1; k < m[i].al; k++)
+			if (words || ends_line(af.list[m[i].a+k])) {
+				if (unmatched)
+					unmatched--;
+				if (changed)
+					changed--;
+				if (extraneous)
+					extraneous--;
+			}
 	}
+	if (!show_wiggles)
+		*wigglesp = wiggles;
+	/* Now count the conflicts and wiggles */
+	for (i = 0; m[i].type != End; i++) {
+		int true_conflict = 0;
+		if (!m[i].in_conflict)
+			continue;
+
+		for (j = i; m[j].type != End && m[j].in_conflict; j++) {
+			if (m[j].in_conflict == 2)
+				true_conflict = 1;
+			if (j > i &&
+			    m[j].in_conflict == 1) {
+				/* end of region */
+				if (!m[j+1].in_conflict)
+					j++;
+				break;
+			}
+		}
+		if (true_conflict)
+			cnt++;
+		else
+			wiggles++;
+		i = j-1;
+	}
+	if (show_wiggles)
+		*wigglesp = wiggles;
 	return cnt;
 }
 
@@ -306,7 +429,8 @@ struct ci make_merger(struct file af, struct file bf, struct file cf,
 	struct ci rv;
 	int i, l;
 	int a, b, c, c1, c2;
-	int wiggle_found = 0;
+	int header_checked = -1;
+	int header_found = 0;
 
 	rv.conflicts = rv.wiggles = rv.ignored = 0;
 
@@ -328,6 +452,17 @@ struct ci make_merger(struct file af, struct file bf, struct file cf,
 		match1 = (a >= csl1[c1].a && b >= csl1[c1].b); /* c1 doesn't match */
 		match2 = (b >= csl2[c2].a && c >= csl2[c2].b);
 
+		if (header_checked != c2) {
+			/* Check if there is a hunk header in this range */
+			int j;
+			header_found = -1;
+			for (j = b; j < csl2[c2].a + csl2[c2].len; j++)
+				if (bf.list[j].start[0] == '\0') {
+					header_found = j;
+					break;
+				}
+			header_checked = c2;
+		}
 		rv.merger[i].a = a;
 		rv.merger[i].b = b;
 		rv.merger[i].c = c;
@@ -337,28 +472,25 @@ struct ci make_merger(struct file af, struct file bf, struct file cf,
 
 		if (!match1 && match2) {
 			/* This is either Unmatched or Extraneous - probably both.
-			 * If the match2 is a hunk-header Extraneous, it must
+			 * If the match2 has a hunk-header Extraneous, it must
 			 * align with an end-of-line in 'a', so adjust endpoint
 			 */
 			int newa = csl1[c1].a;
-			if (b < bf.elcnt && bf.list[b].start
-			    && bf.list[b].start[0] == '\0') {
+			if (header_found >= 0) {
 				while (newa > a &&
 				       !ends_line(af.list[newa-1]))
 					newa--;
-				while (newa < af.elcnt && !(newa == 0 || ends_line(af.list[newa-1])))
-					newa++;
 			}
+			if (a == newa && b == csl1[c1].b)
+				newa = csl1[c1].a;
 			if (a < newa) {
 				/* some unmatched text */
 				rv.merger[i].type = Unmatched;
 				rv.merger[i].al = newa - a;
 				rv.merger[i].bl = 0;
 				rv.merger[i].cl = 0;
-				wiggle_found++;
 			} else {
 				int newb;
-				int j;
 				assert(b < csl1[c1].b);
 				/* some Extraneous text */
 				/* length is min of unmatched on left
@@ -374,18 +506,14 @@ struct ci make_merger(struct file af, struct file bf, struct file cf,
 				newb = b +
 					min(csl1[c1].b - b,
 					    csl2[c2].len - (b-csl2[c2].a));
-				if (bf.list[b].start[0] == '\0')
+				if (header_found == b) {
 					newb = b + 1;
-				for (j = b; j < newb; j++) {
-					if (bf.list[j].start[0] == '\0') {
-						if (wiggle_found > 1)
-							rv.wiggles++;
-						wiggle_found = 0;
-						if (j > b)
-							newb = j;
-					} else
-						wiggle_found++;
+					header_checked = -1;
+				} else if (header_found > b && header_found < newb) {
+					newb = header_found;
+					header_checked = -1;
 				}
+				assert(newb > b);
 				rv.merger[i].cl =
 					rv.merger[i].bl = newb - b;
 			}
@@ -422,7 +550,22 @@ struct ci make_merger(struct file af, struct file bf, struct file cf,
 			if (ignore_already &&
 			    check_alreadyapplied(af, cf, &rv.merger[i]))
 				rv.ignored++;
+			else if (rv.merger[i].bl == 0 &&
+				 rv.merger[i].cl > 0)
+				/* As the 'before' stream is empty, this
+				 * could look like Unmatched in the
+				 * original, and an insertion in the
+				 * diff.  Reporting it like that is
+				 * probably more useful that as a full
+				 * conflict.
+				 * Leave the type for the insertion as
+				 * Conflict (not Changed) as there is some
+				 * real uncertainty here, but allow the
+				 * original to become Unmatched.
+				 */
+				rv.merger[i].al = 0;
 		}
+		rv.merger[i].oldtype = rv.merger[i].type;
 		a += rv.merger[i].al;
 		b += rv.merger[i].bl;
 		c += rv.merger[i].cl;
@@ -440,6 +583,7 @@ struct ci make_merger(struct file af, struct file bf, struct file cf,
 			break;
 	}
 	rv.merger[i].type = End;
+	rv.merger[i].oldtype = End;
 	rv.merger[i].a = a;
 	rv.merger[i].b = b;
 	rv.merger[i].c = c;
@@ -447,31 +591,53 @@ struct ci make_merger(struct file af, struct file bf, struct file cf,
 	rv.merger[i].c2 = c2;
 	rv.merger[i].in_conflict = 0;
 	assert(i < l);
+
+	/* Now revert any AlreadyApplied that aren't bounded by
+	 * Unchanged or Changed.
+	 */
+	for (i = 0; rv.merger[i].type != End; i++) {
+		if (rv.merger[i].type != AlreadyApplied)
+			continue;
+		if (i > 0 && rv.merger[i-1].type != Unchanged &&
+		    rv.merger[i-1].type != Changed)
+			rv.merger[i].type = Conflict;
+		if (rv.merger[i+1].type != Unchanged &&
+		    rv.merger[i+1].type != Changed &&
+		    rv.merger[i+1].type != End)
+			rv.merger[i].type = Conflict;
+	}
 	rv.conflicts = isolate_conflicts(af, bf, cf, csl1, csl2, words,
-					 rv.merger, show_wiggles);
-	if (wiggle_found)
-		rv.wiggles++;
+					 rv.merger, show_wiggles, &rv.wiggles);
 	return rv;
 }
 
-static void printrange(FILE *out, struct file *f, int start, int len)
+static int printrange(FILE *out, struct file *f, int start, int len,
+		      int offset)
 {
+	int lines = 0;
 	while (len > 0) {
-		printword(out, f->list[start]);
+		struct elmnt e = f->list[start];
+		printword(out, e);
+		if (e.start[e.plen-1] == '\n' &&
+		    offset > 0)
+			lines++;
+		offset--;
 		start++;
 		len--;
 	}
+	return lines;
 }
 
-struct ci print_merge2(FILE *out, struct file *a, struct file *b, struct file *c,
-		       struct csl *c1, struct csl *c2,
-		       int words, int ignore_already, int show_wiggles)
+int print_merge(FILE *out, struct file *a, struct file *b, struct file *c,
+		int words, struct merge *merger,
+		struct merge *mpos, int streampos, int offsetpos)
 {
-	struct ci rv = make_merger(*a, *b, *c, c1, c2,
-				   words, ignore_already, show_wiggles);
 	struct merge *m;
+	int lineno = 1;
+	int rv = 0;
+	int offset = INT_MAX;
 
-	for (m = rv.merger; m->type != End ; m++) {
+	for (m = merger; m->type != End ; m++) {
 		struct merge *cm;
 		if (do_trace)
 			printf("[%s: %d-%d,%d-%d,%d-%d%s(%d,%d)]\n",
@@ -493,14 +659,17 @@ struct ci print_merge2(FILE *out, struct file *a, struct file *b, struct file *c
 			 */
 			int found_conflict = 0;
 			int st = 0, st1;
-			if (m->type == Unchanged || m->type == Changed)
-				if (m->hi >= m->lo)
-					st = m->hi;
+			if (m->in_conflict == 1)
+				st = m->hi;
 			st1 = st;
 
-			if (m->type == Unchanged)
-				printrange(out, a, m->a+m->lo, m->hi - m->lo);
+			if (m == mpos)
+				offset = offsetpos;
+			if (m->in_conflict == 1 && m->type == Unchanged)
+				lineno += printrange(out, a, m->a+m->lo, m->hi - m->lo, offset - m->lo);
 
+			if (m == mpos)
+				rv = lineno;
 			if (do_trace)
 				for (cm = m; cm->in_conflict; cm++) {
 					printf("{%s: %d-%d,%d-%d,%d-%d%s(%d,%d)}\n",
@@ -515,71 +684,102 @@ struct ci print_merge2(FILE *out, struct file *a, struct file *b, struct file *c
 					       cm->c, cm->c+cm->cl-1,
 					       cm->in_conflict ? " in_conflict" : "",
 					       cm->lo, cm->hi);
-					if ((cm->type == Unchanged || cm->type == Changed)
-					    && cm != m && cm->lo < cm->hi)
+					if (cm->in_conflict == 1 && cm != m)
 						break;
 				}
 
-			fputs(words ? "<<<---" : "<<<<<<<\n", out);
+			if (m->in_conflict == 1 &&
+			    m[1].in_conflict == 1) {
+				/* Nothing between two conflicts */
+				m++;
+				continue;
+			}
+
+			fputs(words ? "<<<---" : "<<<<<<< found\n", out);
+			if (!words)
+				lineno++;
 			for (cm = m; cm->in_conflict; cm++) {
+				if (cm == mpos && streampos == 0)
+					offset = offsetpos;
 				if (cm->type == Conflict)
 					found_conflict = 1;
-				if ((cm->type == Unchanged || cm->type == Changed)
-				    && cm != m && cm->lo < cm->hi) {
-					printrange(out, a, cm->a, cm->lo);
+				if (cm->in_conflict == 1 && cm != m) {
+					lineno += printrange(out, a, cm->a, cm->lo, offset);
 					break;
 				}
-				printrange(out, a, cm->a+st1, cm->al-st1);
+				lineno += printrange(out, a, cm->a+st1, cm->al-st1, offset-st1);
 				st1 = 0;
+				if (cm == mpos && streampos == 0)
+					rv = lineno;
 			}
-			fputs(words ? "|||" : "|||||||\n", out);
+			if (cm == mpos && streampos == 0)
+				rv = lineno;
+			fputs(words ? "|||" : "||||||| expected\n", out);
+			if (!words)
+				lineno++;
 			st1 = st;
 			for (cm = m; cm->in_conflict; cm++) {
-				if ((cm->type == Unchanged || cm->type == Changed)
-				    && cm != m && cm->lo < cm->hi) {
-					printrange(out, b, cm->b, cm->lo);
+				if (cm == mpos && streampos == 1)
+					offset = offsetpos;
+				if (cm->in_conflict == 1 && cm != m) {
+					lineno += printrange(out, a, cm->a, cm->lo, offset);
 					break;
 				}
-				printrange(out, b, cm->b+st1, cm->bl-st1);
+				lineno += printrange(out, b, cm->b+st1, cm->bl-st1, offset-st1);
 				st1 = 0;
+				if (cm == mpos && streampos == 1)
+					rv = lineno;
 			}
+			if (cm == mpos && streampos == 1)
+				rv = lineno;
 			fputs(words ? "===" : "=======\n", out);
+			if (!words)
+				lineno++;
 			st1 = st;
 			for (cm = m; cm->in_conflict; cm++) {
-				if (cm->type == Unchanged &&
-				    cm != m && cm->lo < cm->hi) {
-					printrange(out, c, cm->c, cm->lo);
+				if (cm == mpos && streampos == 2)
+					offset = offsetpos;
+				if (cm->in_conflict == 1 && cm != m) {
+					if (cm->type == Unchanged)
+						lineno += printrange(out, a, cm->a, cm->lo, offset);
+					else
+						lineno += printrange(out, c, cm->c, cm->cl, offset);
 					break;
 				}
 				if (cm->type == Changed)
 					st1 = 0; /* All of result of change must be printed */
-				printrange(out, c, cm->c+st1, cm->cl-st1);
+				lineno += printrange(out, c, cm->c+st1, cm->cl-st1, offset-st1);
 				st1 = 0;
+				if (cm == mpos && streampos == 2)
+					rv = lineno;
 			}
+			if (cm == mpos && streampos == 2)
+				rv = lineno;
 			if (!found_conflict) {
 				/* This section was wiggled in successfully,
 				 * but full conflict display was requested.
 				 * So now print out the wiggled result as well.
 				 */
-				fputs(words ? "&&&" : "&&&&&&&\n", out);
+				fputs(words ? "&&&" : "&&&&&&& resolution\n", out);
+				if (!words)
+					lineno++;
 				st1 = st;
 				for (cm = m; cm->in_conflict; cm++) {
 					int last = 0;
-					if ((cm->type == Unchanged || cm->type == Changed)
-					    && cm != m && cm->lo < cm->hi)
+					if (cm->in_conflict == 1 && cm != m)
 						last = 1;
 					switch (cm->type) {
 					case Unchanged:
 					case AlreadyApplied:
 					case Unmatched:
-						printrange(out, a, cm->a+st1,
-							   last ? cm->lo : cm->al-st1);
+						lineno += printrange(out, a, cm->a+st1,
+								     last ? cm->lo : cm->al-st1, offset-st1);
 						break;
 					case Extraneous:
 						break;
 					case Changed:
-						printrange(out, c, cm->c+st1,
-							   last ? cm->lo : cm->cl-st1);
+						lineno += printrange(out, c, cm->c,
+								     last ? cm->lo : cm->cl, offset);
 						break;
 					case Conflict:
 					case End:
@@ -590,11 +790,18 @@ struct ci print_merge2(FILE *out, struct file *a, struct file *b, struct file *c
 					st1 = 0;
 				}
 			}
-			fputs(words ? "--->>>" : ">>>>>>>\n", out);
+			fputs(words ? "--->>>" : ">>>>>>> replacement\n", out);
+			if (!words)
+				lineno++;
 			m = cm;
-			if (m->in_conflict && m->type == Unchanged
-			    && m->hi >= m->al) {
-				printrange(out, a, m->a+m->lo, m->hi-m->lo);
+			if (m->in_conflict == 1 && m[1].in_conflict == 0) {
+				/* End of a conflict, no conflict follows */
+				if (m == mpos)
+					offset = offsetpos;
+				if (m->type == Unchanged)
+					lineno += printrange(out, a, m->a+m->lo, m->hi-m->lo, offset-m->lo);
+				if (m == mpos)
+					rv = lineno;
 				m++;
 			}
 		}
@@ -619,22 +826,102 @@ struct ci print_merge2(FILE *out, struct file *a, struct file *b, struct file *c
 			       m->in_conflict ? " in_conflict" : "",
 			       m->lo, m->hi);
 		}
+		if (m == mpos)
+			offset = offsetpos;
 
 		switch (m->type) {
 		case Unchanged:
 		case AlreadyApplied:
 		case Unmatched:
-			printrange(out, a, m->a, m->al);
+			lineno += printrange(out, a, m->a, m->al, offset);
 			break;
 		case Extraneous:
 			break;
 		case Changed:
-			printrange(out, c, m->c, m->cl);
+			lineno += printrange(out, c, m->c, m->cl, offset);
 			break;
 		case Conflict:
 		case End:
 			assert(0);
 		}
+		if (m == mpos)
+			rv = lineno;
 	}
 	return rv;
 }
+
+int save_merge(struct file a, struct file b, struct file c,
+	       struct merge *merger, char *file, int backup)
+{
+	char *replacename = xmalloc(strlen(file) + 20);
+	char *orignew = xmalloc(strlen(file) + 20);
+	int fd;
+	FILE *outfile;
+	int err = 0;
+	int lineno = 0;
+	strcpy(replacename, file);
+	strcat(replacename, "XXXXXX");
+	strcpy(orignew, file);
+	strcat(orignew, ".porig");
+
+	fd = mkstemp(replacename);
+	if (fd < 0) {
+		err = -1;
+		goto out;
+	}
+	outfile = fdopen(fd, "w");
+	lineno = print_merge(outfile, &a, &b, &c, 0, merger,
+			     NULL, 0, 0);
+	fclose(outfile);
+	if (backup && rename(file, orignew) != 0)
+		err = -2;
+	else if (rename(replacename, file) != 0)
+		err = -2;
+
+out:
+	free(replacename);
+	free(orignew);
+	return err < 0 ? err : lineno;
+}
+
+int save_tmp_merge(struct file a, struct file b, struct file c,
+		   struct merge *merger, char **filep,
+		   struct merge *mpos, int streampos, int offsetpos)
+{
+	int fd;
+	FILE *outfile;
+	char *dir, *fname;
+	int lineno;
+	int suffix = 0;
+
+	if (!*filep) {
+		dir = getenv("TMPDIR");
+		if (!dir)
+			dir = "/tmp";
+
+		asprintf(&fname, "%s/wiggle-tmp-XXXXXX", dir);
+	} else {
+		char *base;
+		dir = *filep;
+		base = strrchr(dir, '/');
+		if (base)
+			base++;
+		else
+			base = dir;
+		asprintf(&fname, "%.*stmp-XXXXXX-%s", (int)(base-dir), dir, base);
+		suffix = strlen(base)+1;
+	}
+	fd = mkstemps(fname, suffix);
+
+	if (fd < 0) {
+		free(fname);
+		*filep = NULL;
+		return -1;
+	}
+	outfile = fdopen(fd, "w");
+	lineno = print_merge(outfile, &a, &b, &c, 0, merger,
+			     mpos, streampos, offsetpos);
+	fclose(outfile);
+	*filep = fname;
+	return lineno;
+}
