@@ -2,7 +2,7 @@
  * wiggle - apply rejected patches
  *
  * Copyright (C) 2003 Neil Brown <neilb@cse.unsw.edu.au>
- * Copyright (C) 2010 Neil Brown <neilb@suse.de>
+ * Copyright (C) 2010-2013 Neil Brown <neilb@suse.de>
  *
  *
  *    This program is free software; you can redistribute it and/or modify
@@ -16,8 +16,7 @@
  *    GNU General Public License for more details.
  *
  *    You should have received a copy of the GNU General Public License
- *    along with this program; if not, write to the Free Software Foundation, Inc.,
- *    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
+ *    along with this program.
  *
  *    Author: Neil Brown
  *    Email: <neilb@suse.de>
@@ -89,17 +88,30 @@
 #include	<stdlib.h>
 #include	<stdio.h>
 #include	<ctype.h>
+#include	<sys/stat.h>
 
 char *Cmd = "wiggle";
 int do_trace = 0;
 
-void die()
+void die(char *reason)
 {
-	fprintf(stderr, "%s: fatal error\n", Cmd);
-	abort();
+	fprintf(stderr, "%s: fatal error: %s failure\n", Cmd, reason);
 	exit(3);
 }
 
+void check_dir(char *name, int fd)
+{
+	struct stat st;
+	if (fstat(fd, &st) != 0) {
+		fprintf(stderr, "%s: fatal: %s is strange\n", Cmd, name);
+		exit(3);
+	}
+	if (S_ISDIR(st.st_mode)) {
+		fprintf(stderr, "%s: %s is a directory\n", Cmd, name);
+		exit(3);
+	}
+}
+
 void *xmalloc(int size)
 {
 	void *rv = malloc(size);
@@ -114,11 +126,12 @@ void *xmalloc(int size)
 void printword(FILE *f, struct elmnt e)
 {
 	if (e.start[0])
-		fprintf(f, "%.*s", e.len, e.start);
+		fprintf(f, "%.*s", e.plen + e.prefix,
+			e.start - e.prefix);
 	else {
 		int a, b, c;
 		sscanf(e.start+1, "%d %d %d", &a, &b, &c);
-		fprintf(f, "*** %d,%d **** %d\n", b, c, a);
+		fprintf(f, "*** %d,%d **** %d%s", b, c, a, e.start+18);
 	}
 }
 
@@ -127,7 +140,7 @@ static void printsep(struct elmnt e1, struct elmnt e2)
 	int a, b, c, d, e, f;
 	sscanf(e1.start+1, "%d %d %d", &a, &b, &c);
 	sscanf(e2.start+1, "%d %d %d", &d, &e, &f);
-	printf("@@ -%d,%d +%d,%d @@\n", b, c, e, f);
+	printf("@@ -%d,%d +%d,%d @@%s", b, c, e, f, e1.start+18);
 }
 
 static int extract(int argc, char *argv[], int ispatch, int which)
@@ -383,8 +396,7 @@ static int do_diff(int argc, char *argv[], int obj, int ispatch,
 			f = load_file(argv[1]);
 			if (f.body == NULL) {
 				fprintf(stderr,
-					"%s: cannot load patch"
-					" '%s' - %s\n", Cmd,
+					"%s: cannot load patch '%s' - %s\n", Cmd,
 					argv[1], strerror(errno));
 				return 2;
 			}
@@ -401,8 +413,7 @@ static int do_diff(int argc, char *argv[], int obj, int ispatch,
 			flist[1] = load_file(argv[1]);
 		if (flist[1].body == NULL) {
 			fprintf(stderr,
-				"%s: cannot load file"
-				" '%s' - %s\n", Cmd,
+				"%s: cannot load file '%s' - %s\n", Cmd,
 				argv[1], strerror(errno));
 			return 2;
 		}
@@ -417,13 +428,13 @@ static int do_diff(int argc, char *argv[], int obj, int ispatch,
 		flist[1] = flist[2];
 		flist[2] = f;
 	}
-	fl[0] = split_stream(flist[0], obj == 'l' ? ByLine : ByWord);
-	fl[1] = split_stream(flist[1], obj == 'l' ? ByLine : ByWord);
+	fl[0] = split_stream(flist[0], obj);
+	fl[1] = split_stream(flist[1], obj);
 	if (chunks2 && !chunks1)
 		csl = pdiff(fl[0], fl[1], chunks2);
 	else
-		csl = diff(fl[0], fl[1]);
-	if (obj == 'l') {
+		csl = diff_patch(fl[0], fl[1]);
+	if ((obj & ByMask) == ByLine) {
 		if (!chunks1)
 			printf("@@ -1,%d +1,%d @@\n",
 			       fl[0].elcnt, fl[1].elcnt);
@@ -446,8 +457,9 @@ static int do_diff(int argc, char *argv[], int obj, int ispatch,
 	return exit_status;
 }
 
-static int do_merge(int argc, char *argv[], int obj,
-		    int reverse, int replace, int ignore, int show_wiggles,
+static int do_merge(int argc, char *argv[], int obj, int blanks,
+		    int reverse, int replace, char *outfilename,
+		    int ignore, int show_wiggles,
 		    int quiet)
 {
 	/* merge three files, A B C, so changed between B and C get made to A
@@ -512,7 +524,14 @@ static int do_merge(int argc, char *argv[], int obj,
 			return 2;
 		}
 	}
-	if (replace) {
+	if (outfilename) {
+		outfile = fopen(outfilename, "w");
+		if (!outfile) {
+			fprintf(stderr, "%s: could not create %s\n",
+				Cmd, outfilename);
+			return 2;
+		}
+	} else if (replace) {
 		int fd;
 		replacename = xmalloc(strlen(argv[0]) + 20);
 		orignew = xmalloc(strlen(argv[0]) + 20);
@@ -538,24 +557,24 @@ static int do_merge(int argc, char *argv[], int obj,
 		outfile = fdopen(fd, "w");
 	}
 
-	if (obj == 'l') {
-		fl[0] = split_stream(flist[0], ByLine);
-		fl[1] = split_stream(flist[1], ByLine);
-		fl[2] = split_stream(flist[2], ByLine);
-	} else {
-		fl[0] = split_stream(flist[0], ByWord);
-		fl[1] = split_stream(flist[1], ByWord);
-		fl[2] = split_stream(flist[2], ByWord);
-	}
+	if (obj == 'l')
+		blanks |= ByLine;
+	else
+		blanks |= ByWord;
+	fl[0] = split_stream(flist[0], blanks);
+	fl[1] = split_stream(flist[1], blanks);
+	fl[2] = split_stream(flist[2], blanks);
+
 	if (chunks2 && !chunks1)
 		csl1 = pdiff(fl[0], fl[1], chunks2);
 	else
 		csl1 = diff(fl[0], fl[1]);
-	csl2 = diff(fl[1], fl[2]);
+	csl2 = diff_patch(fl[1], fl[2]);
 
-	ci = print_merge2(outfile, &fl[0], &fl[1], &fl[2],
-			  csl1, csl2, obj == 'w',
-			  ignore, show_wiggles);
+	ci = make_merger(fl[0], fl[1], fl[2], csl1, csl2,
+			 obj == 'w', ignore, show_wiggles > 1);
+	print_merge(outfile, &fl[0], &fl[1], &fl[2],
+		    obj == 'w', ci.merger, NULL, 0, 0);
 	if (!quiet && ci.conflicts)
 		fprintf(stderr,
 			"%d unresolved conflict%s found\n",
@@ -567,7 +586,9 @@ static int do_merge(int argc, char *argv[], int obj,
 			ci.ignored,
 			ci.ignored == 1 ? "" : "s");
 
-	if (replace) {
+	if (outfilename)
+		fclose(outfile);
+	else if (replace) {
 		fclose(outfile);
 		if (rename(argv[0], orignew) == 0 &&
 		    rename(replacename, argv[0]) == 0)
@@ -579,10 +600,13 @@ static int do_merge(int argc, char *argv[], int obj,
 			return 2;
 		}
 	}
-	return (ci.conflicts > 0);
+	if (show_wiggles)
+		return ci.conflicts + ci.wiggles > 0;
+	else
+		return ci.conflicts > 0;
 }
 
-static int multi_merge(int argc, char *argv[], int obj,
+static int multi_merge(int argc, char *argv[], int obj, int blanks,
 		       int reverse, int ignore, int show_wiggles,
 		       int replace, int strip,
 		       int quiet)
@@ -613,6 +637,7 @@ static int multi_merge(int argc, char *argv[], int obj,
 			Cmd, filename);
 		return 2;
 	}
+	check_dir(filename, fileno(f));
 	pl = parse_patch(f, NULL, &num_patches);
 	fclose(f);
 	if (set_prefix(pl, num_patches, strip) == 0) {
@@ -626,7 +651,7 @@ static int multi_merge(int argc, char *argv[], int obj,
 			 pl[i].start, pl[i].end, filename);
 		av[0] = pl[i].file;
 		av[1] = name;
-		rv |= do_merge(2, av, obj, reverse, 1, ignore,
+		rv |= do_merge(2, av, obj, blanks, reverse, 1, NULL, ignore,
 			 show_wiggles, quiet);
 	}
 	return rv;
@@ -649,6 +674,9 @@ int main(int argc, char *argv[])
 	int show_wiggles = 0;
 	char *helpmsg;
 	char *trace;
+	char *outfile = NULL;
+	int selftest = 0;
+	int ignore_blanks = 0;
 
 	trace = getenv("WIGGLE_TRACE");
 	if (trace && *trace)
@@ -694,6 +722,11 @@ int main(int argc, char *argv[])
 				mode = opt;
 				continue;
 			}
+			if (mode == 'B' && opt == 'd') {
+				/* Browse/diff mode */
+				ispatch = 2;
+				continue;
+			}
 			fprintf(stderr,
 				"%s: mode is '%c' - cannot set to '%c'\n",
 				Cmd, mode, opt);
@@ -712,17 +745,28 @@ int main(int argc, char *argv[])
 		case 'r':
 			replace = 1;
 			continue;
+		case 'o':
+			outfile = optarg;
+			replace = 1;
+			continue;
 		case 'R':
 			reverse = 1;
 			continue;
 
+		case 'b':
+			ignore_blanks = IgnoreBlanks;
+			continue;
+
 		case 'i':
 			ignore = 0;
 			continue;
 		case 'W':
-			show_wiggles = 1;
+			show_wiggles = 2;
 			ignore = 0;
 			continue;
+		case REPORT_WIGGLES:
+			show_wiggles = 1;
+			continue;
 
 		case '1':
 		case '2':
@@ -747,13 +791,18 @@ int main(int argc, char *argv[])
 		case 'q':
 			quiet = 1;
 			continue;
+
+		case SELF_TEST:
+			selftest = 1;
+			continue;
 		}
 	if (!mode)
 		mode = 'm';
 
 	if (mode == 'B') {
 		vpatch(argc-optind, argv+optind, ispatch,
-		       strip, reverse, replace);
+		       strip, reverse, replace, outfile, selftest,
+		       ignore_blanks);
 		/* should not return */
 		exit(1);
 	}
@@ -766,9 +815,14 @@ int main(int argc, char *argv[])
 	}
 	if (mode != 'm' && !obj)
 		obj = 'w';
+	if (ispatch && outfile) {
+		fprintf(stderr, "%s: --output incompatible with --patch\n",
+			Cmd);
+		exit(2);
+	}
 	if (replace && mode != 'm') {
 		fprintf(stderr,
-			"%s: --replace only allowed with --merge\n", Cmd);
+			"%s: --replace or --output only allowed with --merge\n", Cmd);
 		exit(2);
 	}
 	if (mode == 'x' && !which) {
@@ -794,19 +848,26 @@ int main(int argc, char *argv[])
 		exit_status = extract(argc-optind, argv+optind, ispatch, which);
 		break;
 	case 'd':
-		exit_status = do_diff(argc-optind, argv+optind, obj, ispatch, which, reverse);
+		exit_status = do_diff(argc-optind, argv+optind,
+				      (obj == 'l' ? ByLine : ByWord)
+				      | ignore_blanks,
+				      ispatch, which, reverse);
 		break;
 	case 'm':
 		if (ispatch)
 			exit_status = multi_merge(argc-optind,
 						  argv+optind, obj,
+						  ignore_blanks,
 						  reverse, ignore,
 						  show_wiggles,
 						  replace, strip,
 						  quiet);
 		else
-			exit_status = do_merge(argc-optind, argv+optind, obj, reverse, replace,
-					       ignore, show_wiggles, quiet);
+			exit_status = do_merge(
+				argc-optind, argv+optind,
+				obj, ignore_blanks, reverse, replace,
+				outfile,
+				ignore, show_wiggles, quiet);
 		break;
 	}
 	exit(exit_status);
